| Copyright | (c) 2019 Athan Clark |
|---|---|
| License | BSD-3-Style |
| Maintainer | athan.clark@gmail.com |
| Portability | GHC |
| Safe Haskell | None |
| Language | Haskell2010 |
Test.Serialization.Symbiote
Description
As an example, say you have some data type TypeA, and some encoding / decoding instance with Aeson
for that data type. Now, you've also got a few functions that work with that data type - f :: TypeA -> TypeA
and g :: TypeA -> TypeA -> TypeA, and you've also taken the time to write a proper Arbitrary instance for TypeA.
Your first order of business in making TypeA a symbiote, is to first demonstrate what operations are supported by it:
instance SymbioteOperation TypeA where
data Operation TypeA
= F
| G TypeA
perform op x = case op of
F -> f x
G y -> g y xYou're also going to need to make sure your new data-family has appropriate serialization instances, as well:
instance ToJSON (Operation TypeA) where
toJSON op = case op of
F -> toJSON "f"
G x -> "g" .: x
instance FromJSON (Operation TypeA) where
parseJSON json = getF <|> getG
where
getF = do
s <- parseJSON json
if s == "f"
then pure F
else fail "Not F"
getG = do
x <- json .: "g"
pure (G x)Next, let's make TypeA an instance of Symbiote:
instance Symbiote TypeA Value where encode = Aeson.toJSON decode = Aeson.parseMaybe Aeson.parseJSON encodeOp = Aeson.toJSON decodeOp = Aeson.parseMaybe Aeson.parseJSON
this instance above actually works for any type that implements ToJSON and FromJSON - there's an orphan
definition in Test.Serialization.Symbiote.Aeson.
Next, you're going to need to actually use this, by registering the type in a test suite:
myFancyTestSuite :: SymbioteT Value IO () myFancyTestSuite = register "TypeA" 100 (Proxy :: Proxy TypeA)
Lastly, you're going to need to actually run the test suite by attaching it to a network. The best way to
do that, is decide whether this peer will be the first or second peer to start the protocol, then use the
respective firstPeer and secondPeer functions - they take as arguments functions that define "how to send"
and "how to receive" messages, and likewise how to report status.
Synopsis
- class SymbioteOperation a where
- class SymbioteOperation a => Symbiote a s where
- newtype EitherOp a = EitherOp (Either a (Operation a))
- data Topic
- type SymbioteT s m = ReaderT Bool (StateT (Map Topic (SymbioteState s)) m)
- register :: forall a s m. Arbitrary a => Arbitrary (Operation a) => Symbiote a s => Eq a => MonadIO m => Topic -> Int -> Proxy a -> SymbioteT s m ()
- firstPeer :: forall m s. MonadIO m => Show s => (First s -> m ()) -> m (Second s) -> (Topic -> m ()) -> (Failure Second s -> m ()) -> (Topic -> Float -> m ()) -> SymbioteT s m () -> m ()
- secondPeer :: forall s m. MonadIO m => Show s => (Second s -> m ()) -> m (First s) -> (Topic -> m ()) -> (Failure First s -> m ()) -> (Topic -> Float -> m ()) -> SymbioteT s m () -> m ()
- data First s
- = AvailableTopics (Map Topic Int)
- | FirstGenerating { }
- | FirstOperating { }
- data Second s
- = BadTopics (Map Topic Int)
- | Start
- | SecondOperating { }
- | SecondGenerating { }
- data Generating s
- = Generated {
- genValue :: s
- genOperation :: s
- | BadResult s
- | YourTurn
- | ImFinished
- | GeneratingNoParseOperated s
- = Generated {
- data Operating s
- data Failure them s
- = BadTopicsFailure { }
- | OutOfSyncFirst (First s)
- | OutOfSyncSecond (Second s)
- | TopicNonexistent Topic
- | WrongTopic { }
- | CantParseOperated Topic s
- | CantParseGeneratedValue Topic s
- | CantParseGeneratedOperation Topic s
- | CantParseLocalValue Topic s
- | CantParseLocalOperation Topic s
- | BadOperating Topic (Operating s)
- | BadGenerating Topic (Generating s)
- | BadThem Topic (them s)
- | SafeFailure {
- safeFailureTopic :: Topic
- safeFailureExpected :: s
- safeFailureGot :: s
- defaultSuccess :: Topic -> IO ()
- defaultFailure :: Show (them s) => Show s => Failure them s -> IO ()
- defaultProgress :: Topic -> Float -> IO ()
- nullProgress :: Topic -> Float -> IO ()
- simpleTest :: MonadBaseControl IO m => MonadIO m => Show s => SymbioteT s m () -> m ()
Documentation
class SymbioteOperation a where Source #
A type-level relation between a type and appropriate, testable operations on that type.
class SymbioteOperation a => Symbiote a s where Source #
A serialization format for a particular type, and serialized data type.
Instances
| (Serialize a, Serialize (Operation a), SymbioteOperation a) => Symbiote a ByteString Source # | |
Defined in Test.Serialization.Symbiote.Cereal.Lazy Methods encode :: a -> ByteString Source # decode :: ByteString -> Maybe a Source # encodeOp :: Operation a -> ByteString Source # | |
| (Serialize a, Serialize (Operation a), SymbioteOperation a) => Symbiote a ByteString Source # | |
Defined in Test.Serialization.Symbiote.Cereal Methods encode :: a -> ByteString Source # decode :: ByteString -> Maybe a Source # encodeOp :: Operation a -> ByteString Source # | |
| (ToJSON a, FromJSON a, ToJSON (Operation a), FromJSON (Operation a), SymbioteOperation a) => Symbiote a Value Source # | |
| SymbioteOperation a => Symbiote a (EitherOp a) Source # | |
The most trivial serialization medium for any a.
Unique name of a type, for a suite of tests
Arguments
| :: Arbitrary a | |
| => Arbitrary (Operation a) | |
| => Symbiote a s | |
| => Eq a | |
| => MonadIO m | |
| => Topic | |
| -> Int | Max size |
| -> Proxy a | |
| -> SymbioteT s m () |
Register a topic in the test suite
Arguments
| :: MonadIO m | |
| => Show s | |
| => (First s -> m ()) | Encode and send first messages |
| -> m (Second s) | Receive and decode second messages |
| -> (Topic -> m ()) | Report when Successful |
| -> (Failure Second s -> m ()) | Report when Failed |
| -> (Topic -> Float -> m ()) | Report on Progress |
| -> SymbioteT s m () | |
| -> m () |
Run the test suite as the first peer
Arguments
| :: MonadIO m | |
| => Show s | |
| => (Second s -> m ()) | Encode and send second messages |
| -> m (First s) | Receive and decode first messages |
| -> (Topic -> m ()) | Report when Successful |
| -> (Failure First s -> m ()) | Report when Failed |
| -> (Topic -> Float -> m ()) | Report on Progress |
| -> SymbioteT s m () | |
| -> m () |
Run the test suite as the second peer
Messages sent by the first peer
Constructors
| AvailableTopics (Map Topic Int) | Mapping of topics to their gen size |
| FirstGenerating | |
Fields | |
| FirstOperating | |
Fields | |
Messages sent by the second peer
Constructors
| BadTopics (Map Topic Int) | |
| Start | |
| SecondOperating | |
Fields | |
| SecondGenerating | |
Fields | |
data Generating s Source #
Messages sent by a peer during their generating phase
Constructors
| Generated | |
Fields
| |
| BadResult s | Expected value |
| YourTurn | |
| ImFinished | |
| GeneratingNoParseOperated s | |
Instances
| Eq s => Eq (Generating s) Source # | |
Defined in Test.Serialization.Symbiote | |
| Show s => Show (Generating s) Source # | |
Defined in Test.Serialization.Symbiote Methods showsPrec :: Int -> Generating s -> ShowS # show :: Generating s -> String # showList :: [Generating s] -> ShowS # | |
Messages sent by a peer during their operating phase
Constructors
| Operated s | Serialized value after operation |
| OperatingNoParseValue s | |
| OperatingNoParseOperation s |
Constructors
| BadTopicsFailure | |
Fields
| |
| OutOfSyncFirst (First s) | |
| OutOfSyncSecond (Second s) | |
| TopicNonexistent Topic | |
| WrongTopic | |
Fields | |
| CantParseOperated Topic s | |
| CantParseGeneratedValue Topic s | |
| CantParseGeneratedOperation Topic s | |
| CantParseLocalValue Topic s | |
| CantParseLocalOperation Topic s | |
| BadOperating Topic (Operating s) | |
| BadGenerating Topic (Generating s) | |
| BadThem Topic (them s) | |
| SafeFailure | |
Fields
| |
defaultSuccess :: Topic -> IO () Source #
Via putStrLn
simpleTest :: MonadBaseControl IO m => MonadIO m => Show s => SymbioteT s m () -> m () Source #
Prints to stdout and uses a local channel for a sanity-check - doesn't serialize.